from typing import List, TypedDict

from typing_extensions import NotRequired

from billing.types._billing_entity import BillingEntity


class Order(BillingEntity):
    pass


class OrderCreatePayload(TypedDict):
    customer_auth_service_id: str
    product_plan_ids: List[str]


class OrderListParams(TypedDict):
    customer_auth_service_id: NotRequired[str]



from datetime import datetime
from typing import Any, ClassVar

from pydantic import VERSION as PYDANTIC_VERSION
from pydantic import BaseModel
from typing_extensions import Self

IS_PYDANTIC_V2 = PYDANTIC_VERSION.startswith("2.")

if IS_PYDANTIC_V2:
    from pydantic import ConfigDict


class BillingObject(BaseModel):
    """
    Data Transfer Object for a billing API object.
    """

    if IS_PYDANTIC_V2:
        model_config: ClassVar[ConfigDict] = ConfigDict(
            frozen=True,
        )
    else:

        class Config:
            frozen = True

    @classmethod
    def parse(cls, obj: Any) -> Self:
        if IS_PYDANTIC_V2:
            return cls.model_validate(obj)
        else:
            return cls.parse_obj(obj)


class BillingEntity(BillingObject):
    id: str
    created_at: datetime


class BillingEntityWithTimestamps(BillingEntity):
    updated_at: datetime



from typing import List

from typing_extensions import Unpack

from billing.services._billing_api_service import BillingAPIService
from billing.types import InvoiceListParams, InvoiceWithItemsBundle


class InvoiceAPIService(BillingAPIService):
    def retrieve(self, object_id: str) -> InvoiceWithItemsBundle:
        return self._request(
            "GET",
            f"/v1/invoices/{object_id}/",
            data_model=InvoiceWithItemsBundle,
        )

    async def retrieve_async(self, object_id: str) -> InvoiceWithItemsBundle:
        return await self._request_async(
            "GET",
            f"/v1/invoices/{object_id}/",
            data_model=InvoiceWithItemsBundle,
        )

    def list(
        self,
        page_number: int = 1,
        page_size: int = 50,
        **params: Unpack[InvoiceListParams],
    ) -> List[InvoiceWithItemsBundle]:
        return self._request(
            "GET",
            "/v1/invoices/",
            params={
                "page": page_number,
                "page_size": page_size,
                **params,
            },
            data_model=InvoiceWithItemsBundle,
            batch_mode=True,
        )

    async def list_async(
        self,
        page_number: int = 1,
        page_size: int = 50,
        **params: Unpack[InvoiceListParams],
    ) -> List[InvoiceWithItemsBundle]:
        return await self._request_async(
            "GET",
            "/v1/invoices/",
            params={
                "page": page_number,
                "page_size": page_size,
                **params,
            },
            data_model=InvoiceWithItemsBundle,
            batch_mode=True,
        )



from decimal import Decimal

from billing.types._billing_entity import BillingEntityWithTimestamps
from billing.types._offer_product_plan import OfferProductPlanWithProduct


class InvoiceItem(BillingEntityWithTimestamps):
    amount: Decimal


class InvoiceItemWithProduct(InvoiceItem):
    offer_product_plan: OfferProductPlanWithProduct



from typing_extensions import Unpack

from billing.services._billing_api_service import BillingAPIService
from billing.types import OfferCancelPayload


class OfferAPIService(BillingAPIService):
    def cancel(
        self,
        object_id: str,
        **payload: Unpack[OfferCancelPayload],
    ) -> None:
        return self._request(
            "POST",
            f"/v1/offers/{object_id}/cancel/",
            json=payload,
        )

    async def cancel_async(
        self,
        object_id: str,
        **payload: Unpack[OfferCancelPayload],
    ) -> None:
        return await self._request_async(
            "POST",
            f"/v1/offers/{object_id}/cancel/",
            json=payload,
        )



from typing import List

from typing_extensions import Unpack

from billing.services._billing_api_service import BillingAPIService
from billing.types import Feature, FeatureRecordPayload, FeatureUsage


class FeatureAPIService(BillingAPIService):
    def retrieve(
        self,
        codename: str,
        customer_id: str,
    ) -> FeatureUsage:
        return self._request(
            "GET",
            f"/v1/features/{codename}/",
            params={"customer_id": customer_id},
            data_model=FeatureUsage,
        )

    async def retrieve_async(
        self,
        codename: str,
        customer_id: str,
    ) -> FeatureUsage:
        return await self._request_async(
            "GET",
            f"/v1/features/{codename}/",
            params={"customer_id": customer_id},
            data_model=FeatureUsage,
        )

    def list(
        self,
        customer_id: str,
        page_number: int = 1,
        page_size: int = 50,
    ) -> List[FeatureUsage]:
        return self._request(
            "GET",
            "/v1/features/",
            params={
                "page": page_number,
                "page_size": page_size,
                "customer_id": customer_id,
            },
            data_model=FeatureUsage,
            batch_mode=True,
        )

    async def list_async(
        self,
        customer_id: str,
        page_number: int = 1,
        page_size: int = 50,
    ) -> List[FeatureUsage]:
        return await self._request_async(
            "GET",
            "/v1/features/",
            params={
                "page": page_number,
                "page_size": page_size,
                "customer_id": customer_id,
            },
            data_model=FeatureUsage,
            batch_mode=True,
        )

    def record(
        self,
        codename: str,
        **payload: Unpack[FeatureRecordPayload],
    ) -> Feature:
        return self._request(
            "POST",
            f"/v1/features/{codename}/",
            json=payload,
            data_model=Feature,
        )

    async def record_async(
        self,
        codename: str,
        **payload: Unpack[FeatureRecordPayload],
    ) -> Feature:
        return await self._request_async(
            "POST",
            f"/v1/features/{codename}/",
            json=payload,
            data_model=Feature,
        )



import asyncio
import random
import time
from typing import Any, Dict, Optional, Union

from httpx import URL, USE_CLIENT_DEFAULT, AsyncClient, Client, Response
from httpx._client import UseClientDefault
from httpx._types import (
    AuthTypes,
    CookieTypes,
    QueryParamTypes,
    RequestContent,
    RequestData,
    RequestFiles,
    TimeoutTypes,
)
from typing_extensions import Unpack

from billing._constants import FEATURE_USAGE_LIMIT_EXCEEDED_ERROR_CODE
from billing._exceptions import AuthenticationError, BillingAPIError, FeatureUsageLimitExceededError, RateLimitError
from billing.types import HTTPXPayload


class HTTPClient:
    MAX_DELAY_WITHOUT_JITTER = 2
    INITIAL_DELAY = 0.5

    def __init__(
        self,
        terminal_secret_key: str,
        terminal_id: str,
        max_network_retries: int = 0,
        setup_sync_client: bool = False,
        setup_async_client: bool = True,
        sync_client: Optional[Client] = None,
        async_client: Optional[AsyncClient] = None,
        **httpx_client_kwargs: Unpack[HTTPXPayload],
    ) -> None:
        if not setup_sync_client and sync_client:
            raise ValueError("Cannot set both `setup_sync_client=False` and `sync_client`.")
        if not setup_async_client and async_client:
            raise ValueError("Cannot set both `setup_sync_client=True` and `async_client`.")
        if not setup_sync_client and not setup_async_client:
            raise ValueError("Either `setup_sync_client=True` or `setup_async_client=True` must be set.")

        self._max_network_retries = max_network_retries

        self._auth_headers = {
            "Authorization": f"Bearer {terminal_secret_key}",
            "X-Terminal-Id": terminal_id,
        }

        if setup_sync_client:
            self._sync_client = sync_client or Client(**httpx_client_kwargs)
        if setup_async_client:
            self._async_client = async_client or AsyncClient(**httpx_client_kwargs)

    def request_with_retries(
        self,
        method: str,
        url: Union[URL, str],
        *,
        content: Optional[RequestContent] = None,
        data: Optional[RequestData] = None,
        files: Optional[RequestFiles] = None,
        json: Optional[Any] = None,
        params: Optional[QueryParamTypes] = None,
        headers: Optional[Dict[str, str]] = None,
        cookies: Optional[CookieTypes] = None,
        auth: Union[AuthTypes, UseClientDefault, None] = USE_CLIENT_DEFAULT,
        follow_redirects: Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
        timeout: Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    ) -> Response:
        attempt_number = 1
        while True:
            try:
                return self._request(
                    method,
                    url,
                    content=content,
                    data=data,
                    files=files,
                    json=json,
                    params=params,
                    headers=headers,
                    cookies=cookies,
                    auth=auth,
                    follow_redirects=follow_redirects,
                    timeout=timeout,
                )
            except BillingAPIError as e:
                if self._is_should_retry(e, attempt_number=attempt_number):
                    time.sleep(self._compute_sleep_seconds(attempt_number))
                else:
                    raise

            attempt_number += 1

    def _request(
        self,
        method: str,
        url: Union[URL, str],
        *,
        content: Optional[RequestContent] = None,
        data: Optional[RequestData] = None,
        files: Optional[RequestFiles] = None,
        json: Optional[Any] = None,
        params: Optional[QueryParamTypes] = None,
        headers: Optional[Dict[str, str]] = None,
        cookies: Optional[CookieTypes] = None,
        auth: Union[AuthTypes, UseClientDefault, None] = USE_CLIENT_DEFAULT,
        follow_redirects: Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
        timeout: Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    ) -> Response:
        if not self._sync_client:
            raise RuntimeError("Sync client is not set.")

        if headers is None:
            headers = {}
        headers.update(self._auth_headers)

        response = self._sync_client.request(
            method,
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
        )

        self._handle_api_error_if_any(response)
        return response

    async def request_async_with_retries(
        self,
        method: str,
        url: Union[URL, str],
        *,
        content: Optional[RequestContent] = None,
        data: Optional[RequestData] = None,
        files: Optional[RequestFiles] = None,
        json: Optional[Any] = None,
        params: Optional[QueryParamTypes] = None,
        headers: Optional[Dict[str, str]] = None,
        cookies: Optional[CookieTypes] = None,
        auth: Union[AuthTypes, UseClientDefault, None] = USE_CLIENT_DEFAULT,
        follow_redirects: Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
        timeout: Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    ) -> Response:

        attempt_number = 1
        while True:
            try:
                return await self._request_async(
                    method,
                    url,
                    content=content,
                    data=data,
                    files=files,
                    json=json,
                    params=params,
                    headers=headers,
                    cookies=cookies,
                    auth=auth,
                    follow_redirects=follow_redirects,
                    timeout=timeout,
                )
            except BillingAPIError as e:
                if self._is_should_retry(e, attempt_number=attempt_number):
                    await asyncio.sleep(self._compute_sleep_seconds(attempt_number))
                else:
                    raise

            attempt_number += 1

    async def _request_async(
        self,
        method: str,
        url: Union[URL, str],
        *,
        content: Optional[RequestContent] = None,
        data: Optional[RequestData] = None,
        files: Optional[RequestFiles] = None,
        json: Optional[Any] = None,
        params: Optional[QueryParamTypes] = None,
        headers: Optional[Dict[str, str]] = None,
        cookies: Optional[CookieTypes] = None,
        auth: Union[AuthTypes, UseClientDefault, None] = USE_CLIENT_DEFAULT,
        follow_redirects: Union[bool, UseClientDefault] = USE_CLIENT_DEFAULT,
        timeout: Union[TimeoutTypes, UseClientDefault] = USE_CLIENT_DEFAULT,
    ) -> Response:
        if not self._async_client:
            raise RuntimeError("Async client is not set.")

        if headers is None:
            headers = {}
        headers.update(self._auth_headers)

        response = await self._async_client.request(
            method,
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=follow_redirects,
            timeout=timeout,
        )

        self._handle_api_error_if_any(response)
        return response

    @staticmethod
    def _handle_api_error_if_any(response: Response) -> None:
        if 200 <= response.status_code < 300:
            return

        if response.status_code in (401, 403):
            raise AuthenticationError(
                status_code=response.status_code,
                response_text=response.text,
            )
        elif response.status_code == 429:
            if FEATURE_USAGE_LIMIT_EXCEEDED_ERROR_CODE in response.text:
                raise FeatureUsageLimitExceededError(
                    status_code=response.status_code,
                    response_text=response.text,
                )

            raise RateLimitError(
                status_code=response.status_code,
                response_text=response.text,
            )

        else:
            raise BillingAPIError(
                status_code=response.status_code,
                response_text=response.text,
            )

    def _is_should_retry(
        self,
        api_error: BillingAPIError,
        attempt_number: int,
    ) -> bool:
        """
        Should retry on conflict, rate limit or server error.
        """

        if attempt_number >= self._max_network_retries + 1:
            return False

        # Should not retry on feature usage limit exceeded error
        if isinstance(api_error, FeatureUsageLimitExceededError):
            return False

        return api_error.status_code in (409, 429) or api_error.status_code >= 500

    @classmethod
    def _compute_sleep_seconds(cls, attempt_number: int) -> float:
        sleep_seconds = min(
            cls.INITIAL_DELAY * (2 ** (attempt_number - 1)),
            cls.MAX_DELAY_WITHOUT_JITTER,
        )

        # Add jitter to the sleep seconds.
        sleep_seconds += random.uniform(0, 1)

        return sleep_seconds  # type: ignore[no-any-return]

    def close_sync_client(self) -> None:
        if self._sync_client:
            self._sync_client.close()

    async def close_async_client(self) -> None:
        if self._async_client:
            await self._async_client.aclose()



from datetime import datetime
from decimal import Decimal
from typing import List, Literal, Optional, TypedDict

from typing_extensions import NotRequired

from billing.types._billing_entity import BillingEntityWithTimestamps
from billing.types._invoice_item import InvoiceItemWithProduct


class Invoice(BillingEntityWithTimestamps):
    total: Decimal
    currency: str
    issued_at: Optional[datetime]
    expired_at: datetime
    paid_at: Optional[datetime]
    refunded_at: Optional[datetime]
    payment_method: Literal["stripe", "telegram_stars"]
    receipt_url: str
    failed_reason: str
    checkout_session_url: Optional[str]


class InvoiceWithItemsBundle(Invoice):
    items: List[InvoiceItemWithProduct]


class InvoiceListParams(TypedDict):
    order_id: NotRequired[str]
    customer_auth_service_id: NotRequired[str]



from billing.types._billing_entity import BillingEntity
from billing.types._product_plan import ProductPlanWithProduct


class OfferProductPlan(BillingEntity):
    offer_id: str
    product_plan_id: str


class OfferProductPlanWithProduct(OfferProductPlan):
    product_plan: ProductPlanWithProduct



from __future__ import annotations

from typing import Any, Literal, Mapping, TypeVar, overload

from httpx import Response

from billing._http_client import HTTPClient
from billing.types import BillingObject, HTTPMethod

_APIObject = TypeVar("_APIObject", bound=BillingObject)


class BillingAPIService:
    def __init__(self, http_client: HTTPClient) -> None:
        self._http_client = http_client

    @overload
    def _request(
        self,
        method: HTTPMethod,
        url: str,
        *,
        json: Any | None = None,
        params: Mapping[str, Any] | None = None,
        data_model: None = None,
        batch_mode: Literal[False] = False,
    ) -> None: ...

    @overload
    def _request(
        self,
        method: HTTPMethod,
        url: str,
        *,
        json: Any | None = None,
        params: Mapping[str, Any] | None = None,
        data_model: type[_APIObject],
        batch_mode: Literal[False] = False,
    ) -> _APIObject: ...

    @overload
    def _request(
        self,
        method: HTTPMethod,
        url: str,
        *,
        json: Any | None = None,
        params: Mapping[str, Any] | None = None,
        data_model: type[_APIObject],
        batch_mode: Literal[True],
    ) -> list[_APIObject]: ...

    def _request(
        self,
        method: HTTPMethod,
        url: str,
        *,
        json: Any | None = None,
        params: Mapping[str, Any] | None = None,
        data_model: type[_APIObject] | None = None,
        batch_mode: bool = False,
    ) -> _APIObject | list[_APIObject] | None:
        response = self._http_client.request_with_retries(
            method,
            url,
            json=json,
            params=params,
        )

        if data_model is None:
            return None

        return self._convert_response_to_entity(
            data_model=data_model,
            batch_mode=batch_mode,
            response=response,
        )

    @overload
    async def _request_async(
        self,
        method: HTTPMethod,
        url: str,
        *,
        json: Any | None = None,
        params: Mapping[str, Any] | None = None,
        data_model: None = None,
        batch_mode: Literal[False] = False,
    ) -> None: ...

    @overload
    async def _request_async(
        self,
        method: HTTPMethod,
        url: str,
        *,
        json: Any | None = None,
        params: Mapping[str, Any] | None = None,
        data_model: type[_APIObject],
        batch_mode: Literal[False] = False,
    ) -> _APIObject: ...

    @overload
    async def _request_async(
        self,
        method: HTTPMethod,
        url: str,
        *,
        json: Any | None = None,
        params: Mapping[str, Any] | None = None,
        data_model: type[_APIObject],
        batch_mode: Literal[True],
    ) -> list[_APIObject]: ...

    async def _request_async(
        self,
        method: HTTPMethod,
        url: str,
        *,
        json: Any | None = None,
        params: Mapping[str, Any] | None = None,
        data_model: type[_APIObject] | None = None,
        batch_mode: bool = False,
    ) -> _APIObject | list[_APIObject] | None:
        response = await self._http_client.request_async_with_retries(
            method,
            url,
            json=json,
            params=params,
        )

        if data_model is None:
            return None

        return self._convert_response_to_entity(
            data_model=data_model,
            batch_mode=batch_mode,
            response=response,
        )

    @staticmethod
    def _convert_response_to_entity(
        *,
        data_model: type[_APIObject],
        batch_mode: bool = False,
        response: Response,
    ) -> _APIObject | list[_APIObject]:
        response_json = response.json()

        if not batch_mode:
            return data_model.parse(response_json)

        if isinstance(response_json, list):
            objects_from_response = response_json
        elif "items" in response_json:
            objects_from_response = response_json["items"]
        else:
            raise ValueError(f"Unexpected response format: {response_json!r}")

        return [data_model.parse(obj) for obj in objects_from_response]



from __future__ import annotations

from typing import Any, Callable, TypedDict

from httpx import URL, Limits
from httpx._types import (
    AuthTypes,
    CertTypes,
    CookieTypes,
    HeaderTypes,
    ProxiesTypes,
    ProxyTypes,
    QueryParamTypes,
    TimeoutTypes,
    VerifyTypes,
)
from typing_extensions import NotRequired


class HTTPXPayload(TypedDict):
    auth: NotRequired[AuthTypes | None]
    params: NotRequired[QueryParamTypes | None]
    headers: NotRequired[HeaderTypes | None]
    cookies: NotRequired[CookieTypes | None]
    verify: NotRequired[VerifyTypes]
    cert: NotRequired[CertTypes | None]
    http1: NotRequired[bool]
    http2: NotRequired[bool]
    proxy: NotRequired[ProxyTypes | None]
    proxies: NotRequired[ProxiesTypes | None]
    timeout: NotRequired[TimeoutTypes]
    follow_redirects: NotRequired[bool]
    limits: NotRequired[Limits]
    max_redirects: NotRequired[int]
    base_url: NotRequired[URL | str]
    app: NotRequired[Callable[..., Any] | None]
    trust_env: NotRequired[bool]
    default_encoding: NotRequired[str | Callable[[bytes], str]]
